import pytest
from pynamodb.attributes import UnicodeAttribute, NumberAttribute
from pynamodb.models import Model

from ..utils import get_key_name


def test_getkeyname_should_raiseerror():
    with pytest.raises(TypeError):
        get_key_name(object)


def test_getkeyname_should_workonstrings():
    class MyModel(Model):
        class Meta:
            table_name = 'some-table'

        notmyid = UnicodeAttribute(null=True)
        myid = UnicodeAttribute(hash_key=True)

    assert get_key_name(MyModel) == 'myid'


def test_getkeyname_should_workonnumbers():
    class MyModel(Model):
        class Meta:
            table_name = 'some-table'

        notmyid = UnicodeAttribute(null=True)
        myid = NumberAttribute(hash_key=True)

    assert get_key_name(MyModel) == 'myid'



from graphene import Dynamic
from graphene import Field
from graphene import ID, Boolean, Int, List, String
from graphene import is_node
from graphene.types.json import JSONString
from pynamodb import attributes
from singledispatch import singledispatch

from graphene_pynamodb import relationships
from graphene_pynamodb.fields import PynamoConnectionField
from graphene_pynamodb.relationships import OneToOne, OneToMany


@singledispatch
def convert_pynamo_attribute(type, attribute, registry=None):
    raise Exception(
        "Don't know how to convert the PynamoDB attribute %s (%s)" % (attribute, attribute.__class__))


@convert_pynamo_attribute.register(attributes.BinaryAttribute)
@convert_pynamo_attribute.register(attributes.UnicodeAttribute)
@convert_pynamo_attribute.register(attributes.UTCDateTimeAttribute)
def convert_column_to_string(type, attribute, registry=None):
    return String(description=getattr(attribute, 'attr_name'),
                  required=not (getattr(attribute, 'null', True)))


@convert_pynamo_attribute.register(relationships.Relationship)
def convert_relationship_to_dynamic(type, attribute, registry=None):
    def dynamic_type():
        _type = registry.get_type_for_model(attribute.model)
        if not _type:
            return None

        if isinstance(attribute, OneToOne):
            return Field(_type)

        if isinstance(attribute, OneToMany):
            if is_node(_type):
                return PynamoConnectionField(_type)
            return Field(List(_type))

    return Dynamic(dynamic_type)


@convert_pynamo_attribute.register(attributes.NumberAttribute)
def convert_column_to_int_or_id(type, attribute, registry=None):
    if attribute.is_hash_key:
        return ID(description=attribute.attr_name, required=not attribute.null)
    else:
        return Int(description=attribute.attr_name, required=not attribute.null)


@convert_pynamo_attribute.register(attributes.BooleanAttribute)
def convert_column_to_boolean(type, attribute, registry=None):
    return Boolean(description=attribute.attr_name, required=not attribute.null)


@convert_pynamo_attribute.register(attributes.UnicodeSetAttribute)
@convert_pynamo_attribute.register(attributes.NumberSetAttribute)
@convert_pynamo_attribute.register(attributes.BinarySetAttribute)
def convert_scalar_list_to_list(type, attribute, registry=None):
    return List(String, description=attribute.attr_name)


@convert_pynamo_attribute.register(attributes.JSONAttribute)
def convert_json_to_string(type, attribute, registry=None):
    return JSONString(description=attribute.attr_name, required=not attribute.null)



from pynamodb.attributes import Attribute
from pynamodb.models import Model

MODEL_KEY_REGISTRY = {}


def get_key_name(model):
    if not issubclass(model, Model):
        raise TypeError("Invalid type passed to get_key_name: %s" % model.__class__)

    if model in MODEL_KEY_REGISTRY:
        return MODEL_KEY_REGISTRY[model]

    for attr in vars(model):
        attr = getattr(model, attr)
        if isinstance(attr, Attribute) and attr.is_hash_key:
            MODEL_KEY_REGISTRY[model] = attr.attr_name
            return attr.attr_name



from __future__ import absolute_import

import os
from datetime import datetime

from pynamodb.attributes import (NumberAttribute, NumberSetAttribute, UnicodeAttribute, UTCDateTimeAttribute)
from pynamodb.models import Model

from graphene_pynamodb.relationships import OneToOne, OneToMany

DB_HOST = None if os.getenv('TRAVIS', False) else "http://localhost:8000"
DB_REGION = "us-west-2"


class Editor(Model):
    class Meta:
        table_name = 'test_graphene_pynamodb_editors'
        host = DB_HOST
        region = DB_REGION

    id = UnicodeAttribute(hash_key=True)
    name = UnicodeAttribute()


class Pet(Model):
    class Meta:
        table_name = 'test_graphene_pynamodb_pets'
        host = DB_HOST
        region = DB_REGION

    id = NumberAttribute(hash_key=True)
    name = UnicodeAttribute()
    reporter_id = NumberAttribute()


class Article(Model):
    class Meta:
        table_name = 'test_graphene_pynamodb_articles'
        host = DB_HOST
        region = DB_REGION

    id = NumberAttribute(hash_key=True)
    headline = UnicodeAttribute()
    pub_date = UTCDateTimeAttribute(default=datetime.now)
    reporter = OneToOne('Reporter')


class Reporter(Model):
    class Meta:
        table_name = 'test_graphene_pynamodb_reporters'
        host = DB_HOST
        region = DB_REGION

    id = NumberAttribute(hash_key=True)
    first_name = UnicodeAttribute()
    last_name = UnicodeAttribute()
    email = UnicodeAttribute(null=True)
    pets = NumberSetAttribute(null=True)
    articles = OneToMany(Article, null=True)
    favorite_article = OneToOne(Article, null=True)



from pynamodb.attributes import Attribute, NumberAttribute
from pynamodb.constants import STRING, ATTR_TYPE_MAP, NUMBER_SHORT
from pynamodb.models import Model
from six import string_types
from wrapt import ObjectProxy

from graphene_pynamodb.utils import get_key_name


class RelationshipResult(ObjectProxy):
    def __init__(self, key_name, key, obj):
        if isinstance(obj, type) and not issubclass(obj, Model):
            raise TypeError("Invalid class passed to RelationshipResult, expected a Model class, got %s" % type(obj))
        super(RelationshipResult, self).__init__(obj)
        self._self_key = key
        self._self_key_name = key_name
        self._self_model = obj

    def __getattr__(self, name):
        if name == self._self_key_name:
            return self._self_key
        if not name.startswith('_') and isinstance(self.__wrapped__, type):
            self.__wrapped__ = self._self_model.get(self._self_key)
        return super(RelationshipResult, self).__getattr__(name)

    def __eq__(self, other):
        return isinstance(other, self._self_model) and self._self_key == getattr(other, self._self_key_name)

    def __ne__(self, other):
        return not self.__eq__(other)


class RelationshipResultList(list):
    def __init__(self, hash_key_name, model, keys):
        self._hash_key_name = hash_key_name
        self._model = model
        self._keys = keys
        super(RelationshipResultList, self).__init__(keys)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return RelationshipResultList(self._hash_key_name, self._model, self._keys[item])

        return RelationshipResult(self._hash_key_name, self._keys[item], self._model)

    def __getslice__(self, i, j):
        return RelationshipResultList(self._hash_key_name, self._model, self._keys[i:j])

    def __iter__(self):
        for key in self._keys:
            yield RelationshipResult(self._hash_key_name, key, self._model)

    def resolve(self):
        models = dict((getattr(entity, self._hash_key_name), entity) for entity in self._model.batch_get(self._keys))
        return [models[key] for key in self._keys]


class Relationship(Attribute):
    _models = None

    @classmethod
    def sub_classes(cls, klass):
        return klass.__subclasses__() + [g for s in klass.__subclasses__() for g in Relationship.sub_classes(s)]

    @classmethod
    def get_model(cls, model_name):
        # Resolve a model name into a model class by looking in all Model subclasses
        if not Relationship._models:
            Relationship._models = Relationship.sub_classes(Model)
        return next((model for model in Relationship._models if model.__name__ == model_name), None)

    def __init__(self, model, lazy=True, **args):
        if not isinstance(model, string_types) and not issubclass(model, Model):
            raise TypeError("Expected PynamoDB Model argument, got: %s " % model.__class__.__name__)

        Attribute.__init__(self, **args)
        self._model = model
        self._lazy = lazy
        self._hash_key_name = None

    @property
    def hash_key_name(self):
        if not self._hash_key_name:
            self._hash_key_name = get_key_name(self.model)
        return self._hash_key_name

    @property
    def model(self):
        if isinstance(self._model, string_types):
            self._model = Relationship.get_model(self._model)

        return self._model


class OneToOne(Relationship):
    attr_type = STRING

    def serialize(self, model):
        return str(getattr(model, self.hash_key_name))

    def deserialize(self, hash_key):
        if isinstance(getattr(self.model, self.hash_key_name), NumberAttribute):
            hash_key = int(hash_key)

        if self._lazy:
            return RelationshipResult(self.hash_key_name, hash_key, self.model)
        else:
            return self.model.get(hash_key)


class OneToMany(Relationship):
    attr_type = 'L'

    def serialize(self, models):
        key_type = ATTR_TYPE_MAP[getattr(self.model, self.hash_key_name).attr_type]
        return [{key_type: str(getattr(model, self.hash_key_name))} for model in models]

    def deserialize(self, hash_keys):
        if hash_keys and isinstance(hash_keys[0], dict):
            key_type = list(hash_keys[0].keys())[0]
            if key_type == NUMBER_SHORT:
                hash_keys = [int(hash_key[key_type]) for hash_key in hash_keys]
            else:
                hash_keys = [hash_key[key_type] for hash_key in hash_keys]
        else:
            if isinstance(getattr(self.model, self.hash_key_name), NumberAttribute):
                hash_keys = [hash_key for hash_key in hash_keys]

        if self._lazy:
            return RelationshipResultList(self.hash_key_name, self.model, hash_keys)
        else:
            return self.model.batch_get(hash_keys)

    def get_value(self, value):
        # we need this for legacy compatibility.
        # deserialize previous string set implementation
        if isinstance(value, dict) and "SS" in value:
            return value["SS"]
        return value["L"]


ATTR_TYPE_MAP["L"] = "L"



import inspect
from collections import OrderedDict

import six
from graphene import Field
from graphene.relay import is_node
from graphene.types.objecttype import ObjectType, ObjectTypeMeta
from graphene.types.options import Options
from graphene.types.utils import merge, yank_fields_from_attrs
from graphene.utils.is_base_type import is_base_type
from pynamodb.attributes import Attribute, NumberAttribute
from pynamodb.models import Model

from .converter import convert_pynamo_attribute
from .registry import Registry, get_global_registry
from .relationships import RelationshipResult
from .utils import get_key_name


def get_model_fields(model):
    attributes = dict()
    for attr_name in vars(model):
        attr = getattr(model, attr_name)
        if isinstance(attr, Attribute):
            attributes[attr_name] = attr

    return OrderedDict(sorted(attributes.items(), key=lambda t: t[0]))


def construct_fields(options):
    only_fields = options.only_fields
    exclude_fields = options.exclude_fields
    inspected_model = get_model_fields(options.model)

    fields = OrderedDict()

    for name, attribute in inspected_model.items():
        is_not_in_only = only_fields and name not in only_fields
        is_already_created = name in options.fields
        is_excluded = name in exclude_fields or is_already_created
        if is_not_in_only or is_excluded:
            # We skip this field if we specify only_fields and is not
            # in there. Or when we excldue this field in exclude_fields
            continue
        converted_column = convert_pynamo_attribute(attribute, attribute, options.registry)
        fields[name] = converted_column

    return fields


class PynamoObjectTypeMeta(ObjectTypeMeta):
    @staticmethod
    def __new__(cls, name, bases, attrs):
        # Also ensure initialization is only performed for subclasses of Model
        # (excluding Model class itself).
        if not is_base_type(bases, PynamoObjectTypeMeta):
            return type.__new__(cls, name, bases, attrs)

        options = Options(
            attrs.pop('Meta', None),
            name=name,
            description=attrs.pop('__doc__', None),
            model=None,
            local_fields=None,
            only_fields=(),
            exclude_fields=(),
            id='id',
            interfaces=(),
            registry=None
        )

        if not options.registry:
            options.registry = get_global_registry()
        assert isinstance(options.registry, Registry), (
            'The attribute registry in {}.Meta needs to be an'
            ' instance of Registry, received "{}".'
        ).format(name, options.registry)

        assert (inspect.isclass(options.model) and issubclass(options.model, Model)), (
            'You need to pass a valid PynamoDB Model in '
            '{}.Meta, received "{}".'
        ).format(name, options.model)

        cls = ObjectTypeMeta.__new__(cls, name, bases, dict(attrs, _meta=options))

        options.registry.register(cls)

        options.pynamo_fields = yank_fields_from_attrs(
            construct_fields(options),
            _as=Field,
        )
        options.fields = merge(
            options.interface_fields,
            options.pynamo_fields,
            options.base_fields,
            options.local_fields
        )

        return cls


class PynamoObjectType(six.with_metaclass(PynamoObjectTypeMeta, ObjectType)):
    @classmethod
    def is_type_of(cls, root, context, info):
        if isinstance(root, RelationshipResult) and root.__wrapped__ == cls._meta.model:
            return True
        return isinstance(root, cls._meta.model)

    @classmethod
    def get_node(cls, id, context, info):
        if isinstance(getattr(cls._meta.model, get_key_name(cls._meta.model)), NumberAttribute):
            return cls._meta.model.get(int(id))
        else:
            return cls._meta.model.get(id)

    def resolve_id(self, args, context, info):
        graphene_type = info.parent_type.graphene_type
        if is_node(graphene_type):
            return getattr(self, get_key_name(graphene_type._meta.model))



from setuptools import find_packages, setup

setup(
    name='graphene-pynamodb',
    version='0.8.0',

    description='Graphene PynamoDB integration',
    long_description=open('README.rst').read(),

    url='https://github.com/yfilali/graphql-pynamodb',

    author='Yacine Filali',
    author_email='yfilali@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],

    keywords='api graphql protocol rest relay graphene pynamodb dynamodb',

    packages=find_packages(exclude=['tests']),

    install_requires=[
        'six>=1.10.0',
        'graphene>=1.0',
        'pynamodb>=1.6.0',
        'singledispatch>=3.4.0.3',
        'wrapt>=1.10.8'
    ],
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest>=2.7.2',
        'mock'
    ],
    test_suite="graphene_pynamodb.tests",

)
